
#ifndef H_PENGUINSOCCER
 #define H_PENGUINSOCCER

#include <assert.h>
#include <string>

/* The board is conceptually rotated so that the
 * "normally lower left" corner (0,0) is SOUTH.
 * Therefore NORTHWEST is "up" (+0,+1), and
 * NORTHEAST is "right" (+1,+0). */
enum Direction {
    SOUTH, SOUTHWEST, WEST,
    SOUTHEAST, NO_DIRECTION, NORTHWEST,
    EAST, NORTHEAST, NORTH
};

static inline Direction& operator++ (Direction &d)
{ d = (Direction)((int)d + 1); return d; }

Direction turn_towards(Direction d, Direction target);

struct Pos {
    int x, y;

    Pos() { }
    Pos(int x_, int y_): x(x_), y(y_) { }

    Pos & operator += (Direction d);

    bool operator == (const Pos &rhs) const
    { return (x == rhs.x && y == rhs.y); }

    bool operator != (const Pos &rhs) const
    { return (x != rhs.x || y != rhs.y); }

    int manhattan_distance(const Pos &rhs) const;

    bool off_board() const;
    bool isGoal(int who) const
    { return (who==1 && x==0 && y==0) || (who==0 && x==7 && y==7); }
    bool isHome(int who) const
    { return (who==0 && x==0 && y==0) || (who==1 && x==7 && y==7); }
    bool isCenter() const
    { return (x==3 || x==4) && (y==3 || y==4); }
};

static inline Pos operator + (const Pos &p, Direction d)
{
    Pos q = p;
    return q += d;
}

struct Penguin {
    int color;
    int size;

    bool on_board;
    Pos pos;
    bool standing;
    Direction dir;

    std::string toString() const;
};

struct Board {
    int attacker;
    Penguin penguin[2][3];  /* {0,1}{BABY, PAPA, MAMA} */
    bool ball_on_board;
    Pos ball_pos;

    Board();
    bool has_ball(const Penguin &) const;
    Penguin *penguinAt(const Pos &);
    const Penguin *penguinAt(const Pos &) const;
    bool game_over() const;

    void display() const;
    std::string toComparableString() const;
};
int findattacker(const Board &);

enum MoveKind { SLIDE, STAND, KICK };

struct Move {
    MoveKind kind;  /* STAND, SLIDE, KICK */
    int penguin_id;  /* i.e., BABY, PAPA, or MAMA */
    Direction dir;  /* SLIDE, KICK */
    Direction final_dir;  /* SLIDE */

    Move() { }
    Move(MoveKind mk, int i, Direction d, Direction fd):
        kind(SLIDE), penguin_id(i), dir(d), final_dir(fd)
        { assert(mk == SLIDE); }
    Move(MoveKind mk, int i, Direction d):
        kind(mk), penguin_id(i), dir(d)
        { assert(mk == SLIDE || mk == KICK); }
    Move(MoveKind mk, int i):
        kind(STAND), penguin_id(i)
        { assert(mk == STAND); }

    std::string toString(int who) const;
    bool scan(const char *, const Board &);
};

bool is_valid_move(const Board &b, const Move &m);
bool apply_move(Board &b, const Move &m);

#endif /* H_PENGUINSOCCER */
